@@ -1,51 +1,48 @@
 import logging
-from typing import Optional, Union, Literal, NoReturn, TYPE_CHECKING
+from typing import Optional, Union, NoReturn, Any
 from dataclasses import dataclass
+from threading import Thread, Lock
 import time
 
 from cachetools import TTLCache
 import cherrypy
 
-if TYPE_CHECKING:
-    from telebox.telegram_bot.telegram_bot import TelegramBot
+from telebox.telegram_bot.telegram_bot import TelegramBot
 from telebox.telegram_bot.types.types.update import Update
 from telebox.telegram_bot.types.types.message import Message
 from telebox.telegram_bot.types.types.callback_query import CallbackQuery
-from telebox.dispatcher.event_queue import EventQueue
+from telebox.telegram_bot.errors import UnknownUpdateContentError, UnknownMessageContentError
+from telebox.dispatcher.typing import Event
+from telebox.dispatcher.media_group import MediaGroup
+from telebox.dispatcher.event_queue import EventQueue, Event as Event_
 from telebox.dispatcher.enums.event_type import EventType
 from telebox.dispatcher.handlers.handlers.event import AbstractEventHandler
 from telebox.dispatcher.handlers.handlers.error import AbstractErrorHandler
-from telebox.dispatcher.filters.filter import (
-    AbstractFilter,
-    ExpressionType,
-    AbstractExpression,
-    NoneExpression,
-    FilterExpression,
-    InvertExpression,
-    AndExpression,
-    OrExpression
-)
-from telebox.dispatcher.filters.event_filter import AbstractEventFilter
-from telebox.dispatcher.filters.error_filter import AbstractErrorFilter
+from telebox.dispatcher.filters.event_filter import AbstractEventBaseFilter
+from telebox.dispatcher.filters.error_filter import AbstractErrorBaseFilter
+from telebox.dispatcher.filters.event.none import NoneFilter
+from telebox.dispatcher.filters.error.none import ErrorNoneFilter
 from telebox.dispatcher.middlewares.middleware import Middleware
 from telebox.dispatcher.rate_limiter import RateLimiter
 from telebox.dispatcher.server_root import ServerRoot
 from telebox.dispatcher.errors import DispatcherError
 from telebox.utils.thread_pool import ThreadPool
 from telebox.utils.not_set import NotSet
-from telebox.utils.request_timeout import RequestTimeout
-from telebox.typing import Event
-from telebox.utils.context.vars import (
+from telebox.context.vars import (
     event_context,
     event_handler_context,
     error_handler_context
 )
 
 
+logger = logging.getLogger(__name__)
+_MEDIA_GROUP_GATHERING_DELAY_SECS = 0.1
+
+
 @dataclass
 class EventHandler:
     handler: AbstractEventHandler
-    filter: AbstractExpression
+    filter: AbstractEventBaseFilter
     rate_limiter: Optional[RateLimiter] = None
     calls: Optional[TTLCache] = None
 
@@ -53,7 +50,7 @@ class EventHandler:
 @dataclass
 class ErrorHandler:
     handler: AbstractErrorHandler
-    filter: AbstractExpression
+    filter: AbstractErrorBaseFilter
 
 
 @dataclass
@@ -61,16 +58,22 @@ class CallState:
     is_first: bool
 
 
-logger = logging.getLogger(__name__)
-EventFilter = Union[AbstractEventFilter, AbstractExpression, None]
-ErrorFilter = Union[AbstractErrorFilter, AbstractExpression, None]
+class TimeContainer:
+
+    def __init__(self, initial_item: Any):
+        self.items = [initial_item]
+        self.time = time.monotonic()
+
+    def add(self, item: Any) -> None:
+        self.items.append(item)
+        self.time = time.monotonic()
 
 
 class Dispatcher:
 
     def __init__(
         self,
-        bot: "TelegramBot",
+        bot: TelegramBot,
         *,
         drop_over_limit_events: bool = False,
         default_rate_limiters: Optional[dict[EventType, RateLimiter]] = None
@@ -85,11 +88,14 @@ class Dispatcher:
         self._event_handlers: dict[EventType, list[EventHandler]] = {i: [] for i in EventType}
         self._error_handlers: list[ErrorHandler] = []
         self._middlewares: list[Middleware] = []
+        self._media_group_gathering_thread: Optional[Thread] = None
+        self._media_group_messages: dict[str, TimeContainer] = {}
+        self._media_group_message_lock = Lock()
 
     def add_message_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.MESSAGE, filter_, rate_limiter)
@@ -97,29 +103,37 @@ class Dispatcher:
     def add_edited_message_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.EDITED_MESSAGE, filter_, rate_limiter)
 
+    def add_media_group_handler(
+        self,
+        handler: AbstractEventHandler,
+        filter_: Optional[AbstractEventBaseFilter] = None,
+        rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
+    ) -> None:
+        self._add_event_handler(handler, EventType.MEDIA_GROUP, filter_, rate_limiter)
+
     def add_channel_post_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None
+        filter_: Optional[AbstractEventBaseFilter] = None
     ) -> None:
         self._add_event_handler(handler, EventType.CHANNEL_POST, filter_, None)
 
     def add_edited_channel_post_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None
+        filter_: Optional[AbstractEventBaseFilter] = None
     ) -> None:
         self._add_event_handler(handler, EventType.EDITED_CHANNEL_POST, filter_, None)
 
     def add_inline_query_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.INLINE_QUERY, filter_, rate_limiter)
@@ -127,7 +141,7 @@ class Dispatcher:
     def add_chosen_inline_result_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.CHOSEN_INLINE_RESULT, filter_, rate_limiter)
@@ -135,7 +149,7 @@ class Dispatcher:
     def add_callback_query_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.CALLBACK_QUERY, filter_, rate_limiter)
@@ -143,7 +157,7 @@ class Dispatcher:
     def add_shipping_query_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.SHIPPING_QUERY, filter_, rate_limiter)
@@ -151,7 +165,7 @@ class Dispatcher:
     def add_pre_checkout_query_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.PRE_CHECKOUT_QUERY, filter_, rate_limiter)
@@ -159,14 +173,14 @@ class Dispatcher:
     def add_poll_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None
+        filter_: Optional[AbstractEventBaseFilter] = None
     ) -> None:
         self._add_event_handler(handler, EventType.POLL, filter_, None)
 
     def add_poll_answer_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
         self._add_event_handler(handler, EventType.POLL_ANSWER, filter_, rate_limiter)
@@ -174,33 +188,36 @@ class Dispatcher:
     def add_my_chat_member_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None
+        filter_: Optional[AbstractEventBaseFilter] = None
     ) -> None:
         self._add_event_handler(handler, EventType.MY_CHAT_MEMBER, filter_, None)
 
     def add_chat_member_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None
+        filter_: Optional[AbstractEventBaseFilter] = None
     ) -> None:
         self._add_event_handler(handler, EventType.CHAT_MEMBER, filter_, None)
 
     def add_chat_join_request_handler(
         self,
         handler: AbstractEventHandler,
-        filter_: EventFilter = None
+        filter_: Optional[AbstractEventBaseFilter] = None
     ) -> None:
         self._add_event_handler(handler, EventType.CHAT_JOIN_REQUEST, filter_, None)
 
     def add_error_handler(
         self,
         handler: AbstractErrorHandler,
-        filter_: ErrorFilter = None
+        filter_: Optional[AbstractErrorBaseFilter] = None
     ) -> None:
+        if filter_ is None:
+            filter_ = ErrorNoneFilter()
+
         self._error_handlers.append(
             ErrorHandler(
                 handler=handler,
-                filter=_get_expression(filter_)
+                filter=filter_
             )
         )
 
@@ -211,7 +228,7 @@ class Dispatcher:
         self,
         threads: int,
         *,
-        request_timeout: Optional[RequestTimeout] = None,
+        timeout_secs: Union[int, float, None] = None,
         delay_secs: Union[int, float] = 0.2,
         error_delay_secs: Union[int, float] = 5.0,
         limit: Optional[int] = None,
@@ -233,6 +250,7 @@ class Dispatcher:
         self._polling_is_used = True
         logger.debug("Polling is starting...")
         offset_update_id = None
+        self._run_media_group_gathering_thread()
         self._run_thread_pool(threads)
         logger.info("Polling started.")
 
@@ -241,7 +259,7 @@ class Dispatcher:
                 # noinspection PyBroadException
                 try:
                     updates = self._bot.get_updates(
-                        request_timeout=request_timeout,
+                        timeout_secs=timeout_secs,
                         offset=offset_update_id,
                         limit=limit,
                         timeout=timeout,
@@ -261,12 +279,14 @@ class Dispatcher:
         except KeyboardInterrupt:
             logger.info("Polling stopped.")
             self._finish_update_processing()
+            self._media_group_gathering_thread = None
+            self._thread_pool = None
             self._polling_is_used = False
 
     def run_server(
         self,
         threads: int,
-        webhook_path: str = "",
+        webhook_path: Optional[str] = None,
         *,
         host: str = "0.0.0.0",
         port: int = 443,
@@ -295,38 +315,41 @@ class Dispatcher:
                 "server.ssl_private_key": ssl_private_key_path,
             })
 
+        self._run_media_group_gathering_thread()
         self._run_thread_pool(threads)
         logger.info("Server started.")
 
         try:
             cherrypy.quickstart(
                 root=ServerRoot(self._process_update),
-                script_name=webhook_path
+                script_name=webhook_path.rstrip("/") if webhook_path else str()
             )
         finally:
             logger.info("Server stopped.")
             self._finish_update_processing()
+            self._media_group_gathering_thread = None
+            self._thread_pool = None
             self._server_is_used = False
 
     def drop_pending_updates(
         self,
         *,
-        request_timeout: Optional[RequestTimeout] = None,
+        timeout_secs: Union[int, float, None] = None,
         with_delete_webhook: bool = True
     ) -> None:
         logger.debug("Dropping pending updates...")
 
         if with_delete_webhook:
-            self._bot.delete_webhook(request_timeout=request_timeout, drop_pending_updates=True)
+            self._bot.delete_webhook(timeout_secs=timeout_secs, drop_pending_updates=True)
         else:
             updates = self._bot.get_updates(
-                request_timeout=request_timeout,
+                timeout_secs=timeout_secs,
                 offset=-1
             )
 
             if updates:
                 self._bot.get_updates(
-                    request_timeout=request_timeout,
+                    timeout_secs=timeout_secs,
                     offset=updates[-1].update_id + 1
                 )
 
@@ -336,9 +359,15 @@ class Dispatcher:
         self,
         handler: AbstractEventHandler,
         event_type: EventType,
-        filter_: EventFilter = None,
+        filter_: Optional[AbstractEventBaseFilter] = None,
         rate_limiter: Union[RateLimiter, None, NotSet] = NotSet()
     ) -> None:
+        if filter_ is None:
+            filter_ = NoneFilter()
+
+        if not filter_.check_event_type(event_type):
+            raise DispatcherError(f"{event_type!r} is not supported by this filter!")
+
         if rate_limiter is NotSet():
             rate_limiter = self._default_rate_limiters.get(event_type)
 
@@ -350,7 +379,7 @@ class Dispatcher:
         self._event_handlers[event_type].append(
             EventHandler(
                 handler=handler,
-                filter=_get_expression(filter_),
+                filter=filter_,
                 rate_limiter=rate_limiter,
                 calls=calls
             )
@@ -361,14 +390,30 @@ class Dispatcher:
         event: Event,
         event_type: EventType
     ) -> Optional[EventHandler]:
-        return _get_handler(self._event_handlers[event_type], (event,))
+        values = {}
+
+        for i in self._event_handlers[event_type]:
+            if (i.filter is None) or i.filter.get_result(event, event_type, values):
+                return i
 
     def _get_error_handler(
         self,
         error: Exception,
-        event: Event
+        event: Event,
+        event_type: EventType
     ) -> Optional[ErrorHandler]:
-        return _get_handler(self._error_handlers, (error, event))
+        values = {}
+
+        for i in self._error_handlers:
+            if (i.filter is None) or i.filter.get_result(error, event, event_type, values):
+                return i
+
+    def _run_media_group_gathering_thread(self) -> None:
+        self._media_group_gathering_thread = Thread(
+            target=self._run_media_group_gathering,
+            daemon=True
+        )
+        self._media_group_gathering_thread.start()
 
     def _run_thread_pool(self, threads: int) -> None:
         self._thread_pool = ThreadPool(
@@ -380,24 +425,73 @@ class Dispatcher:
 
     def _process_update(self, update: Update) -> None:
         logger.debug("Update received: %r.", update)
-        event, event_type = update.content
 
-        if (
-            self._drop_over_limit_events
-            and isinstance(event, (Message, CallbackQuery))
-            and event.chat_id in self._bot.over_limit_chat_ids
-        ):
+        try:
+            event, content_type = update.content
+        except UnknownUpdateContentError:
+            logger.debug(
+                "Update skipped because it contains an unknown content type: %r.",
+                update
+            )
+
+            return
+
+        event_type = EventType(content_type.value)
+
+        if isinstance(event, Message):
+            try:
+                event.content  # noqa
+            except UnknownMessageContentError:
+                logger.debug(
+                    "Event skipped because it contains an unknown content type: %r.",
+                    event
+                )
+
+                return
+
+            if event.media_group_id is not None:
+                with self._media_group_message_lock:
+                    if event.media_group_id in self._media_group_messages:
+                        self._media_group_messages[event.media_group_id].add(event)
+                    else:
+                        self._media_group_messages[event.media_group_id] = TimeContainer(event)
+
+                return
+
+        if self._check_over_limit_event(event):
             logger.debug("Event from over limit chat dropped: %r.", event)
         else:
             logger.debug("Event added to queue: %r.", event)
             self._events.add_event(event, event_type)
 
+    def _check_over_limit_event(self, event: Event) -> bool:
+        return (
+            self._drop_over_limit_events
+            and isinstance(event, (Message, CallbackQuery, MediaGroup))
+            and event.chat_id in self._bot.over_limit_chat_ids
+        )
+
     def _finish_update_processing(self) -> None:
         logger.info("Update processing finishing...")
         self._events.wait_events()
-        self._thread_pool = None
         logger.info("Update processing finished.")
 
+    def _run_media_group_gathering(self) -> NoReturn:
+        while True:
+            with self._media_group_message_lock:
+                for i in tuple(self._media_group_messages):
+                    if (time.monotonic() - self._media_group_messages[i].time) > 0.1:
+                        event = MediaGroup(self._media_group_messages[i].items)
+                        del self._media_group_messages[i]
+
+                        if self._check_over_limit_event(event):
+                            logger.debug("Event from over limit chat dropped: %r.", event)
+                        else:
+                            logger.debug("Event added to queue: %r.", event)
+                            self._events.add_event(event, EventType.MEDIA_GROUP)
+
+            time.sleep(_MEDIA_GROUP_GATHERING_DELAY_SECS)
+
     def _run_event_processing(self, events: EventQueue) -> NoReturn:
         while True:
             event_context_token = event_handler_context_token = error_handler_context_token = None
@@ -411,7 +505,10 @@ class Dispatcher:
                 for i in self._middlewares:
                     i.pre_process_event(event.event, event.event_type)
 
-                event_handler = self._get_event_handler(event.event, event.event_type)
+                event_handler = self._get_event_handler(
+                    event=event.event,
+                    event_type=event.event_type
+                )
 
                 if event_handler is not None:
                     event_handler_context_token = event_handler_context.set(event_handler.handler)
@@ -428,7 +525,11 @@ class Dispatcher:
                         for i in self._middlewares:
                             i.pre_process_error(error, event.event, event.event_type)
 
-                        error_handler = self._get_error_handler(error, event.event)
+                        error_handler = self._get_error_handler(
+                            error=error,
+                            event=event.event,
+                            event_type=event.event_type
+                        )
 
                         if error_handler is None:
                             raise
@@ -466,7 +567,7 @@ class Dispatcher:
                 logger.debug("Event processing finished: %r.", event)
 
 
-def _process_rate_limiting(event: Event, handler: EventHandler) -> bool:
+def _process_rate_limiting(event: Event_, handler: EventHandler) -> bool:
     if handler.rate_limiter is not None:
         if event.user_id not in handler.calls:
             handler.calls[event.user_id] = CallState(is_first=True)
@@ -478,90 +579,3 @@ def _process_rate_limiting(event: Event, handler: EventHandler) -> bool:
             return True
 
     return False
-
-
-def _get_handler(handlers: list, value_getting_args: tuple):
-    filter_values = {}
-
-    for i in handlers:
-        result_getter = _EXPRESSION_RESULT_GETTERS[i.filter.get_type()]
-
-        if result_getter(i.filter, filter_values, value_getting_args):
-            return i
-
-
-# noinspection PyUnusedLocal
-def _get_none_expression_result(
-    expression: NoneExpression,
-    values: dict,
-    value_getting_args: tuple
-) -> Literal[True]:
-    return True
-
-
-def _get_filter_expression_result(
-    expression: FilterExpression,
-    values: dict,
-    value_getting_args: tuple
-) -> bool:
-    filter_ = expression.value
-    filter_type = type(filter_)
-
-    try:
-        value = values[filter_type]
-    except KeyError:
-        value = values[filter_type] = filter_.get_value(*value_getting_args)
-
-    return filter_.check_value(value)
-
-
-def _get_invert_expression_result(
-    expression: InvertExpression,
-    values: dict,
-    value_getting_args: tuple
-) -> bool:
-    result_getter = _EXPRESSION_RESULT_GETTERS[expression.value.get_type()]
-
-    return not result_getter(expression.value, values, value_getting_args)
-
-
-def _get_and_expression_result(
-    expression: AndExpression,
-    values: dict,
-    value_getting_args: tuple
-) -> bool:
-    return all(
-        _EXPRESSION_RESULT_GETTERS[i.get_type()](i, values, value_getting_args)
-        for i in expression
-    )
-
-
-def _get_or_expression_result(
-    expression: OrExpression,
-    values: dict,
-    value_getting_args: tuple
-) -> bool:
-    return any(
-        _EXPRESSION_RESULT_GETTERS[i.get_type()](i, values, value_getting_args)
-        for i in expression
-    )
-
-
-_EXPRESSION_RESULT_GETTERS = {
-    ExpressionType.NONE: _get_none_expression_result,
-    ExpressionType.FILTER: _get_filter_expression_result,
-    ExpressionType.INVERT: _get_invert_expression_result,
-    ExpressionType.AND: _get_and_expression_result,
-    ExpressionType.OR: _get_or_expression_result
-}
-
-
-def _get_expression(filter_) -> AbstractExpression:
-    if filter_ is None:
-        return NoneExpression()
-    elif isinstance(filter_, AbstractFilter):
-        return FilterExpression(filter_)
-    elif isinstance(filter_, AbstractExpression):
-        return filter_
-
-    raise ValueError(f"Unknown filter type {filter_!r}!")
