@@ -5,4 +5,4 @@ from telebox.errors import TeleboxError
 
 @dataclass
 class DispatcherError(TeleboxError):
-    """General class for dispatcher errors."""
+    """Class for dispatcher errors."""
